use crate::prelude::*;

use anyhow::Result;

use std::fmt;

#[derive(Clone)]
pub enum Address {
    DeviceAddress(VkDeviceAddress),
    HostAddressMut(VkDeviceOrHostAddressKHR),
    HostAddressConst(VkDeviceOrHostAddressConstKHR),
}

impl Address {
    #[inline]
    pub fn device_address_mut(&self) -> Result<VkDeviceOrHostAddressKHR> {
        match self {
            Self::DeviceAddress(address) => Ok(VkDeviceOrHostAddressKHR::from(*address)),
            Self::HostAddressMut(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: DeviceAddress found: HostAddressMut",
            )),
            Self::HostAddressConst(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: DeviceAddress found: HostAddressConst",
            )),
        }
    }

    #[inline]
    pub fn device_address_const(&self) -> Result<VkDeviceOrHostAddressConstKHR> {
        match self {
            Self::DeviceAddress(address) => Ok(VkDeviceOrHostAddressConstKHR::from(*address)),
            Self::HostAddressMut(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: DeviceAddress found: HostAddressMut",
            )),
            Self::HostAddressConst(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: DeviceAddress found: HostAddressConst",
            )),
        }
    }

    #[inline]
    pub fn host_address_mut(&self) -> Result<VkDeviceOrHostAddressKHR> {
        match self {
            Self::DeviceAddress(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: HostAddressMut found: DeviceAddress",
            )),
            Self::HostAddressMut(address) => Ok(*address),
            Self::HostAddressConst(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: HostAddressMut found: HostAddressConst",
            )),
        }
    }

    #[inline]
    pub fn host_address_const(&self) -> Result<VkDeviceOrHostAddressConstKHR> {
        match self {
            Self::DeviceAddress(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: HostAddressConst found: DeviceAddress",
            )),
            Self::HostAddressMut(_) => Err(anyhow::Error::msg(
                "Wrong device address format. Expected: HostAddressConst found: HostAddressMut",
            )),
            Self::HostAddressConst(address) => Ok(*address),
        }
    }

    #[inline]
    pub fn is_device_address(&self) -> bool {
        match self {
            Self::DeviceAddress(_) => true,
            _ => false,
        }
    }

    #[inline]
    pub fn is_host_address(&self) -> bool {
        match self {
            Self::DeviceAddress(_) => false,
            Self::HostAddressConst(_) => true,
            Self::HostAddressMut(_) => true,
        }
    }

    #[inline]
    pub fn is_host_address_mut(&self) -> bool {
        match self {
            Self::DeviceAddress(_) => false,
            Self::HostAddressConst(_) => false,
            Self::HostAddressMut(_) => true,
        }
    }

    #[inline]
    pub fn is_host_address_const(&self) -> bool {
        match self {
            Self::DeviceAddress(_) => false,
            Self::HostAddressConst(_) => true,
            Self::HostAddressMut(_) => false,
        }
    }
}

impl<'a, T> From<&'a mut T> for Address {
    fn from(reference: &'a mut T) -> Self {
        Self::HostAddressMut(VkDeviceOrHostAddressKHR::from(reference))
    }
}

impl<'a, T> From<&'a T> for Address {
    fn from(reference: &'a T) -> Self {
        Self::HostAddressConst(VkDeviceOrHostAddressConstKHR::from(reference))
    }
}

impl From<VkDeviceAddress> for Address {
    fn from(device_address: VkDeviceAddress) -> Self {
        Self::DeviceAddress(device_address)
    }
}

impl Into<VkDeviceOrHostAddressKHR> for Address {
    fn into(self) -> VkDeviceOrHostAddressKHR {
        match self {
            Self::DeviceAddress(address) => VkDeviceOrHostAddressKHR::from(address),
            Self::HostAddressMut(address) => address,
            Self::HostAddressConst(_) => {
                panic!("Called Into<VkDeviceOrHostAddressKHR> on HostAddressConst")
            }
        }
    }
}

impl Into<VkDeviceOrHostAddressConstKHR> for Address {
    fn into(self) -> VkDeviceOrHostAddressConstKHR {
        match self {
            Self::DeviceAddress(address) => VkDeviceOrHostAddressConstKHR::from(address),
            Self::HostAddressMut(address) => address.into(),
            Self::HostAddressConst(address) => address,
        }
    }
}

impl Into<VkDeviceAddress> for Address {
    fn into(self) -> VkDeviceAddress {
        match self {
            Self::DeviceAddress(address) => address,
            Self::HostAddressMut(address) => address.device_address(),
            Self::HostAddressConst(address) => address.device_address(),
        }
    }
}

impl fmt::Debug for Address {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("Address").finish()
    }
}
